@@ -5,41 +5,39 @@ see templateexporter.py.
 # Copyright (c) Jupyter Development Team.
 # Distributed under the terms of the Modified BSD License.
 
-from __future__ import print_function, absolute_import
 
-import io
-import os
-import copy
 import collections
+import copy
 import datetime
+import os
 import sys
+from typing import Optional
 
 import nbformat
-
-from traitlets.config.configurable import LoggingConfigurable
+from nbformat import validator
+from traitlets import Bool, HasTraits, List, TraitError, Unicode
 from traitlets.config import Config
-from traitlets import Bool, HasTraits, Unicode, List, TraitError
+from traitlets.config.configurable import LoggingConfigurable
 from traitlets.utils.importstring import import_item
-from typing import Optional
 
 
 class ResourcesDict(collections.defaultdict):
     def __missing__(self, key):
-        return ''
+        return ""
 
 
 class FilenameExtension(Unicode):
     """A trait for filename extensions."""
 
-    default_value = u''
-    info_text = 'a filename extension, beginning with a dot'
+    default_value = ""
+    info_text = "a filename extension, beginning with a dot"
 
     def validate(self, obj, value):
         # cast to proper unicode
         value = super().validate(obj, value)
 
         # check that it starts with a dot
-        if value and not value.startswith('.'):
+        if value and not value.startswith("."):
             msg = "FileExtension trait '{}' does not begin with a dot: {!r}"
             raise TraitError(msg.format(self.name, value))
 
@@ -53,44 +51,50 @@ class Exporter(LoggingConfigurable):
     accompanying resources dict.
     """
 
-    enabled = Bool(True,
-        help = "Disable this exporter (and any exporters inherited from it)."
-    ).tag(config=True)
+    enabled = Bool(True, help="Disable this exporter (and any exporters inherited from it).").tag(
+        config=True
+    )
 
     file_extension = FilenameExtension(
         help="Extension of the file that should be written to disk"
     ).tag(config=True)
 
+    optimistic_validation = Bool(
+        False,
+        help="Reduces the number of validation steps so that it only occurs after all preprocesors have run.",
+    ).tag(config=True)
+
     # MIME type of the result file, for HTTP response headers.
     # This is *not* a traitlet, because we want to be able to access it from
     # the class, not just on instances.
-    output_mimetype = ''
+    output_mimetype = ""
 
     # Should this converter be accessible from the notebook front-end?
     # If so, should be a friendly name to display (and possibly translated).
     export_from_notebook = None
 
-    #Configurability, allows the user to easily add filters and preprocessors.
-    preprocessors = List(
-        help="""List of preprocessors, by name or namespace, to enable."""
-    ).tag(config=True)
+    # Configurability, allows the user to easily add filters and preprocessors.
+    preprocessors = List(help="""List of preprocessors, by name or namespace, to enable.""").tag(
+        config=True
+    )
 
     _preprocessors = List()
 
-    default_preprocessors = List([
-                                  'nbconvert.preprocessors.TagRemovePreprocessor',
-                                  'nbconvert.preprocessors.RegexRemovePreprocessor',
-                                  'nbconvert.preprocessors.ClearOutputPreprocessor',
-                                  'nbconvert.preprocessors.ExecutePreprocessor',
-                                  'nbconvert.preprocessors.coalesce_streams',
-                                  'nbconvert.preprocessors.SVG2PDFPreprocessor',
-                                  'nbconvert.preprocessors.LatexPreprocessor',
-                                  'nbconvert.preprocessors.HighlightMagicsPreprocessor',
-                                  'nbconvert.preprocessors.ExtractOutputPreprocessor',
-                                  'nbconvert.preprocessors.ClearMetadataPreprocessor',
-                              ],
+    default_preprocessors = List(
+        [
+            "nbconvert.preprocessors.TagRemovePreprocessor",
+            "nbconvert.preprocessors.RegexRemovePreprocessor",
+            "nbconvert.preprocessors.ClearOutputPreprocessor",
+            "nbconvert.preprocessors.ExecutePreprocessor",
+            "nbconvert.preprocessors.coalesce_streams",
+            "nbconvert.preprocessors.SVG2PDFPreprocessor",
+            "nbconvert.preprocessors.LatexPreprocessor",
+            "nbconvert.preprocessors.HighlightMagicsPreprocessor",
+            "nbconvert.preprocessors.ExtractOutputPreprocessor",
+            "nbconvert.preprocessors.ClearMetadataPreprocessor",
+        ],
         help="""List of preprocessors available by default, by name, namespace,
-        instance, or type."""
+        instance, or type.""",
     ).tag(config=True)
 
     def __init__(self, config=None, **kw):
@@ -112,7 +116,7 @@ class Exporter(LoggingConfigurable):
         super().__init__(config=with_default_config, **kw)
 
         self._init_preprocessors()
-
+        self._nb_metadata = {}
 
     @property
     def default_config(self):
@@ -124,7 +128,7 @@ class Exporter(LoggingConfigurable):
 
         Parameters
         ----------
-        nb : `nbformat.NotebookNode`
+        nb : :class:`~nbformat.NotebookNode`
             Notebook node (dict-like with attr-access)
         resources : dict
             Additional resources that can be accessed read/write by
@@ -136,12 +140,17 @@ class Exporter(LoggingConfigurable):
         nb_copy = copy.deepcopy(nb)
         resources = self._init_resources(resources)
 
-        if 'language' in nb['metadata']:
-            resources['language'] = nb['metadata']['language'].lower()
+        if "language" in nb["metadata"]:
+            resources["language"] = nb["metadata"]["language"].lower()
 
         # Preprocess
         nb_copy, resources = self._preprocess(nb_copy, resources)
-
+        notebook_name = ""
+        if resources is not None:
+            name = resources.get("metadata", {}).get("name", "")
+            path = resources.get("metadata", {}).get("path", "")
+            notebook_name = os.path.join(path, name)
+        self._nb_metadata[notebook_name] = nb_copy.metadata
         return nb_copy, resources
 
     def from_filename(self, filename: str, resources: Optional[dict] = None, **kw):
@@ -162,25 +171,24 @@ class Exporter(LoggingConfigurable):
         # Pull the metadata from the filesystem.
         if resources is None:
             resources = ResourcesDict()
-        if not 'metadata' in resources or resources['metadata'] == '':
-            resources['metadata'] = ResourcesDict()
+        if "metadata" not in resources or resources["metadata"] == "":
+            resources["metadata"] = ResourcesDict()
         path, basename = os.path.split(filename)
         notebook_name = os.path.splitext(basename)[0]
-        resources['metadata']['name'] = notebook_name
-        resources['metadata']['path'] = path
+        resources["metadata"]["name"] = notebook_name
+        resources["metadata"]["path"] = path
 
         modified_date = datetime.datetime.fromtimestamp(os.path.getmtime(filename))
         # datetime.strftime date format for ipython
-        if sys.platform == 'win32':
+        if sys.platform == "win32":
             date_format = "%B %d, %Y"
         else:
             date_format = "%B %-d, %Y"
-        resources['metadata']['modified_date'] = modified_date.strftime(date_format)
+        resources["metadata"]["modified_date"] = modified_date.strftime(date_format)
 
-        with io.open(filename, encoding='utf-8') as f:
+        with open(filename, encoding="utf-8") as f:
             return self.from_file(f, resources=resources, **kw)
 
-
     def from_file(self, file_stream, resources=None, **kw):
         """
         Convert a notebook from a notebook file.
@@ -196,8 +204,9 @@ class Exporter(LoggingConfigurable):
             Ignored
 
         """
-        return self.from_notebook_node(nbformat.read(file_stream, as_version=4), resources=resources, **kw)
-
+        return self.from_notebook_node(
+            nbformat.read(file_stream, as_version=4), resources=resources, **kw
+        )
 
     def register_preprocessor(self, preprocessor, enabled=False):
         """
@@ -216,18 +225,21 @@ class Exporter(LoggingConfigurable):
 
         """
         if preprocessor is None:
-            raise TypeError('preprocessor must not be None')
+            raise TypeError("preprocessor must not be None")
         isclass = isinstance(preprocessor, type)
         constructed = not isclass
 
         # Handle preprocessor's registration based on it's type
-        if constructed and isinstance(preprocessor, str,):
+        if constructed and isinstance(
+            preprocessor,
+            str,
+        ):
             # Preprocessor is a string, import the namespace and recursively call
             # this register_preprocessor method
             preprocessor_cls = import_item(preprocessor)
             return self.register_preprocessor(preprocessor_cls, enabled)
 
-        if constructed and hasattr(preprocessor, '__call__'):
+        if constructed and hasattr(preprocessor, "__call__"):  # noqa
             # Preprocessor is a function, no need to construct it.
             # Register and return the preprocessor.
             if enabled:
@@ -247,8 +259,9 @@ class Exporter(LoggingConfigurable):
         else:
             # Preprocessor is an instance of something without a __call__
             # attribute.
-            raise TypeError('preprocessor must be callable or an importable constructor, got %r' % preprocessor)
-
+            raise TypeError(
+                "preprocessor must be callable or an importable constructor, got %r" % preprocessor
+            )
 
     def _init_preprocessors(self):
         """
@@ -265,10 +278,9 @@ class Exporter(LoggingConfigurable):
         for preprocessor in self.preprocessors:
             self.register_preprocessor(preprocessor, enabled=True)
 
-
     def _init_resources(self, resources):
 
-        #Make sure the resources dict is of ResourcesDict type.
+        # Make sure the resources dict is of ResourcesDict type.
         if resources is None:
             resources = ResourcesDict()
         if not isinstance(resources, ResourcesDict):
@@ -276,21 +288,30 @@ class Exporter(LoggingConfigurable):
             new_resources.update(resources)
             resources = new_resources
 
-        #Make sure the metadata extension exists in resources
-        if 'metadata' in resources:
-            if not isinstance(resources['metadata'], ResourcesDict):
+        # Make sure the metadata extension exists in resources
+        if "metadata" in resources:
+            if not isinstance(resources["metadata"], ResourcesDict):
                 new_metadata = ResourcesDict()
-                new_metadata.update(resources['metadata'])
-                resources['metadata'] = new_metadata
+                new_metadata.update(resources["metadata"])
+                resources["metadata"] = new_metadata
         else:
-            resources['metadata'] = ResourcesDict()
-            if not resources['metadata']['name']:
-                resources['metadata']['name'] = 'Notebook'
+            resources["metadata"] = ResourcesDict()
+            if not resources["metadata"]["name"]:
+                resources["metadata"]["name"] = "Notebook"
 
-        #Set the output extension
-        resources['output_extension'] = self.file_extension
+        # Set the output extension
+        resources["output_extension"] = self.file_extension
         return resources
 
+    def _validate_preprocessor(self, nbc, preprocessor):
+        try:
+            if not hasattr(validator, "normalize"):
+                nbformat.validate(nbc, relax_add_props=True)
+            else:
+                nbformat.validate(nbc)
+        except nbformat.ValidationError:
+            self.log.error("Notebook is invalid after preprocessor %s", preprocessor)
+            raise
 
     def _preprocess(self, nb, resources):
         """
@@ -309,18 +330,20 @@ class Exporter(LoggingConfigurable):
 
         # Do a copy.deepcopy first,
         # we are never safe enough with what the preprocessors could do.
-        nbc =  copy.deepcopy(nb)
+        nbc = copy.deepcopy(nb)
         resc = copy.deepcopy(resources)
 
+        if hasattr(validator, "normalize"):
+            _, nbc = validator.normalize(nbc)
+
         # Run each preprocessor on the notebook.  Carry the output along
         # to each preprocessor
         for preprocessor in self._preprocessors:
             nbc, resc = preprocessor(nbc, resc)
-            try:
-                nbformat.validate(nbc, relax_add_props=True)
-            except nbformat.ValidationError:
-                self.log.error('Notebook is invalid after preprocessor %s',
-                               preprocessor)
-                raise
+            if not self.optimistic_validation:
+                self._validate_preprocessor(nbc, preprocessor)
+
+        if self.optimistic_validation:
+            self._validate_preprocessor(nbc, preprocessor)
 
         return nbc, resc
